;**************************************************************************
;* A P P L I C A T I O N   N O T E   F O R   T H E   A V R   F A M I L Y
;*
;* Number		:AVR????
;* Title		:Software SPI Master 
;* Date			:19.06.97
;* Support telephone	:+47 72 88 87 20 (Atmel Norway)
;* Support fax		:+47 72 88 87 18 (Atmel Norway)
;* Target MCU		:AT90Sxxxx
;*
;* Code Size		:82 words
;* Low Register Usage	:1
;* High Register Usage	:4
;* Interrupt Usage	:Timer/Counter0 overflow interrupt
;*
;* DESCRIPTION
;*
;* This application note describes how to make a software Master SPI interface
;* on any AVR device with the 8-bit Timer/Counter0.
;*
;* The constants P and N determine the data rate. P selects the clock prescale
;* as described in the T/C Prescaler in the AVR databook. If the T/C prescaling
;* is denoted C, the following expression yields the data rate:
;*
;*				         XTAL
;*			 BIT-RATE ~ ----------------
;*				     2*C*N + (11/C)
;*
;* Absolute minimum value for N*C is 27 (which causes the interruptflag to be 
;* set again before the interrupt is finished). A lower value will cause the
;* timer register to wrap around zero and some bits will in the transmission
;* will use very long time to complete. The bitrate is not exact since we don't
;* exactly know how long it takes before the interrupt routine is serviced.
;*
;* A typical transmit sequence is as follows:
;*
;*	1.  Set SPI settings
;*	2.  Enable global int's (sei)
;*	3.  Enable SS. If slave requires some delay time beween SS
;*	    activiation and data trasnsmission, the user must wait
;*	    here.
;*	4.  Set s_data
;*	5.  Call spi_transmit
;*	6.  Do other things...
;*	7.  Wait for busy to clear
;*	8.  Read s_data
;*	9a. If cpha=0: Disable SS and possibly wait some time (the slave must
;*	    be able to detect the deactiviation before it is activated again)
;*          Goto 3.
;*	9b. If cpha=1: Goto 4
;*
;**************************************************************************
.include "1200def.inc"


;***** Bitrate settings

.equ	N=27			;Bitrate (minimum value is 27, lower value 
				;  gives no better performance)
.equ	P=1			;Prescale value
.equ	C=1			;Clock divisor


;***** SPI Global Registers

.def	s_sr		=r15	;Status-register storage
.def	s_tmp		=r16	;Scratchregister
.def	s_tmp2		=r17	;Scratchregister 2
.def	s_status	=r18	;Status-register
.def	s_data		=r19	;Data to transmit (data recieved)


;***** Bit positions in the Status-register

.equ	cpol	=7		;Clock polarity
.equ	cpha	=6		;Clock phase
.equ	busy	=5		;Busy flag (Read only)
.equ	bc4	=4		;Bit 4-0 are used as edge counter
.equ	bc3	=3		;  while trasmitting. Read Only
.equ	bc2	=2
.equ	bc1	=1
.equ	bc0	=0


;***** Bits i/o on portd

.equ	mosi	=0
.equ	miso	=1
.equ	ss	=4
.equ	sck	=2


;***** General definitions

.equ	msb	=7
.equ	lsb	=0



;**************************************************************************
;*
;*	PROGRAM START - EXECUTION STARTS HERE
;*
;**************************************************************************
	.cseg

	.org $0000
	rjmp	start			;Reset handler

	.org OVF0addr
	rjmp	tim0_ovf		;Timer0 overflow handler



;**************************************************************************
;*
;* TIM0_OVF - Timer/Counter 0 Overflow Interrupt
;*
;*
;* DESCRIPTION
;* This is the main transmission routine. It is invoked (automatically) when 
;* SCK should toggle. It samples and/or drives MOSI/MISO depending on
;* is's state. When done, it clears the busyflag and stops itself.
;*
;*
;* Total number of words	: 32 
;* Total number of cycles	: 24 (incl. reti) - min. 19, max. 32
;* Low register usage		: 1 (s_sr)
;* High register usage		: 4 (s_tmp,s_tmp2,s_status,s_data)
;*
;**************************************************************************
tim0_ovf:
	in	s_sr,sreg			;Store status-register

	in	s_tmp,tcnt0			;Set reload value (compeanste
	subi	s_tmp,n				;  against already used cycles)
	out	tcnt0,s_tmp

	inc	s_status			;Increase bit counter
	sbrc	s_status,bc4			;Jump if counter is = 16 (finished)
	rjmp	spi_last

spi_toggle:
	in	s_tmp,portd			;Toggle sck. Read portd
	ldi	s_tmp2,(1<<sck)			;Load bit to toggle
	eor	s_tmp,s_tmp2			;Excl. or with portd
	out	portd,s_tmp			;Put the port data back

	sbrs	s_status,lsb			;Jump if it is even cycles
	rjmp	spi_even

spi_odd:					;Sample miso @ odd cycles
	sec					;Set carry
	sbis	pind,miso			;Clear carry if MISO
	clc					;  is cleared
	rol	s_data				;Rotate data left
	rjmp	spi_end

spi_even:					;Set bit on mosi @ even cycles
	sbrc	s_data,msb			;Set bit on MOSI if
	sbi	portd,mosi			;  msb is set
	sbrs	s_data,msb			;Clear bit on MOSI if
	cbi	portd,mosi			;  msb is cleared

spi_end:
	out	sreg,s_sr			;Restore statusregister and exit
	reti

spi_last:					;Last bit is sent. Quitting routine
	in	s_tmp,timsk			;Read timer register int. enable register
	cbr	s_tmp,(1<<toie0)		;Disable the timer interrupt
	out	timsk,s_tmp			;Write the register back

	andi	s_status,(1<<cpol)|(1<<cpha)	;Delete all bits but cpol and cpha

	sbrs	s_status,cpha			;If cpha is cleared
	rjmp	spi_toggle			;  Jump to spi_toggle 

	out	sreg,s_sr			;Restore statusregister and exit
	reti



;**************************************************************************
;*
;* SPI_TRANSMIT - Initiates and trasmits the data in s_data
;*
;*
;* DESCRIPTION
;* This routine initiates the exchange of data between the master and the
;* slave. It enables the timer/counter interrupt to transmit the data 
;* async. Data to be sent is located in s_data. This routine sets the 
;* busy flag in s_status. A new byte is ready to be transmitted when
;* this flag is cleared.
;*
;* NOTE: Modify only the cpol and cpha flags when busy is not set. This
;* prevents errors in the trasmission. s_data must NOT be altered while
;* the busy-flag is set.
;*
;*
;* Total number of words	: 28
;* Total number of cycles	: 33 (incl. ret)
;* High register usage		: 3 (s_status,s_data,s_tmp)
;*
;**************************************************************************
spi_transmit:

;Set status - delete all bits but cpol and cpha and set busy flag
	andi	s_status,(1<<cpol)|(1<<cpha)
	sbr	s_status,(1<<busy)

;Set the correct SCK as a function of cpol and cpha
	clr	s_tmp				;Clear register
	sbrc	s_status,cpol			;If cpol is set:
	inc	s_tmp				;  Increase register
	sbrc	s_status,cpha			;If cpha is set:
	inc	s_tmp				;  Increase register
	sbrc	s_tmp,lsb			;Set sck equal to lsb (Set high
	sbi	portd,sck			;  only when either cpol
	sbrs	s_tmp,lsb			;  or cpha is set, but not
	cbi	portd,sck			;  both - xor)

;Set mosi output to msb in data
	sbrc	s_data,msb			;If msb in s_data is set
	sbi	portd,mosi			;  Set MOSI high
	sbrs	s_data,msb			;If msb in s_data is not set
	cbi	portd,mosi			;  Set MOSI low

;Erase pending timer interrupts
	in	s_tmp,tifr			;Get current timer value
	sbr	s_tmp,(1<<tov0)			;Set bit in register (to erase the flag)
	out	tifr,s_tmp			;Write back the modified value

;Enable timer/counter 0 interrupts
	in	s_tmp,timsk			;Get old register value
	sbr	s_tmp,(1<<toie0)		;Modify the register (enable int)
	out	timsk,s_tmp			;Write the register back

;Set the timer's prescale (and thus start the timer)
	in	s_tmp,tccr0			;Get old value
	cbr	s_tmp,(1<<cs02)|(1<<cs01)|(1<<cs00)	;Clear old settings
	ori	s_tmp,p				;Set new values
	out	tccr0,s_tmp			;Write out the register

;Set reload value
	ldi	s_tmp,(256-n)
	out	tcnt0,s_tmp

	ret



;**************************************************************************
;*
;*	Test/Example Program
;*
;* This example program is a demostration of a typical master transmission.
;* Port b is scanned and it's contents is transmitted to the slave.
;*	
;**************************************************************************
.def	tmp	=r21			;Temporary variable


start:
	ldi	s_status,(1<<cpha)		;Set SPI mode (1<<cpha) and/or (1<<cpol)

	ldi	tmp,(1<<mosi)|(1<<ss)|(1<<sck)	;Set port output direction and pins
	out	portd,tmp
	out	ddrd,tmp

	ser	tmp				;Set portb output (to enable keypad on
	out	portb,tmp			;  the dev. board)

	sei					;Enable global interrupt


send:	cbi	portd,ss			;Enable slave select

	ldi	tmp,$40				;Wait some time (some SPI devices - like the
wait:	dec	tmp				;  software slave SPI - must have som time
	brne	wait				;  between SS activation and the first
						;  clock edge - SCK )

	;=== SET S_DATA (OUT) HERE ===
	in	s_data,pinb			;Read pinb

	rcall	spi_transmit			;Initiate transmission

	;=== DO SOME OTHER STUFF HERE ===

loop:	sbrc	s_status,busy			;Wait for busy flag to clear
	rjmp	loop

	;=== READ S_DATA (IN) HERE ===

	sbi	portd,ss			;Disable slave select

	ldi	tmp,$40				;Wait some time. (The ss signal must
wait2:	dec	tmp				;  be high long enough so the slave
	brne	wait2				;  can detect it )

	rjmp	send				;Repeat...

